pub mod dm_type;
pub mod dvm_value;
pub mod lib;
pub mod mem;
mod object;
pub mod object_type;
pub mod op_codes;
pub mod platform;
mod pointer;
pub mod util;

use crate::vm::dvm_value::DvmValue;
use crate::vm::lib::{DmConstant, DmLib};
use crate::vm::object::DmAllocObject;
use crate::vm::object_type::DmFn;
use op_codes::*;
use std::alloc::{alloc, dealloc, Layout};
use std::collections::HashMap;
use std::rc::Rc;

pub type PlatformFunction =
    fn(args: Vec<DvmValue>, state: &mut DvmState, context: &DvmContext) -> DvmValue;

#[derive(Debug)]
enum CallFrame {
    PlatformCall(CallFrameInfo),
    DeimosCall(CallFrameInfo),
}

#[derive(Debug)]
struct CallFrameInfo {
    pub fqn: String,
}

pub struct DvmContext {
    libs: Vec<Rc<DmLib>>,
    functions: HashMap<String, Rc<DmFn>>,
    platform_functions: HashMap<String, Rc<PlatformFunction>>,
}

impl DvmContext {
    pub fn new() -> Self {
        DvmContext {
            libs: Vec::new(),
            functions: HashMap::new(),
            platform_functions: HashMap::new(),
        }
    }

    pub fn load_libs(&mut self, libs: Vec<Rc<DmLib>>) {
        for lib in libs {
            self.libs.push(lib.clone());
            for lib_fn in &lib.functions {
                self.functions
                    .insert(lib_fn.fqn().to_string(), lib_fn.clone());
            }
            for interface_function in lib
                .interfaces
                .iter()
                .flat_map(|interface| interface.get_functions())
            {
                self.functions.insert(
                    interface_function.fqn().to_string(),
                    interface_function.clone(),
                );
            }
            for implementation_function in lib
                .implementations
                .iter()
                .flat_map(|implementation| &implementation.functions)
            {
                self.functions.insert(
                    implementation_function.fqn().to_string(),
                    implementation_function.clone(),
                );
            }
            for method in lib
                .implementations
                .iter()
                .flat_map(|implementation| &implementation.methods)
            {
                self.functions
                    .insert(method.dm_fn().fqn().to_string(), method.dm_fn().clone());
            }
        }
    }

    pub fn load_platform_fns(&mut self, platform_fns: &HashMap<String, PlatformFunction>) {
        for (fqn, platform_fn) in platform_fns {
            self.platform_functions
                .insert(fqn.to_string(), Rc::new(*platform_fn));
        }
    }

    pub fn fn_by_fqn(&self, fqn: &str) -> Option<Rc<DmFn>> {
        self.functions.get(fqn).cloned()
    }

    pub fn platform_fn_by_fqn(&self, fqn: &str) -> Option<Rc<PlatformFunction>> {
        self.platform_functions.get(fqn).cloned()
    }
}

pub struct DvmState {
    call_stack: Vec<CallFrame>,
    registers: Vec<DvmValue>,
    register_state_stack: Vec<Vec<DvmValue>>,
}

impl DvmState {
    pub fn new() -> Self {
        DvmState {
            call_stack: Vec::new(),
            registers: Vec::new(),
            register_state_stack: Vec::new(),
        }
    }
}

macro_rules! dump_state {
    ( $message: expr, $state: expr ) => {
        println!("----");
        println!("{}", $message);
        println!("----");

        println!("Call stack:");
        for call_frame in &$state.call_stack {
            match call_frame {
                CallFrame::PlatformCall(frame_info) => {
                    println!("  {}", frame_info.fqn);
                }
                CallFrame::DeimosCall(frame_info) => {
                    println!("  {}", frame_info.fqn);
                }
            }
        }
        println!("Registers: ");
        for (i, register) in $state.registers.iter().enumerate() {
            println!("  r{}: {:?}", i, register);
        }
    };
}

pub fn call_fn(
    state: &mut DvmState,
    context: &DvmContext,
    dm_fn: &DmFn,
    args: Vec<DvmValue>,
) -> DvmValue {
    // save current state
    state.call_stack.push(CallFrame::DeimosCall(CallFrameInfo {
        fqn: dm_fn.fqn().to_string(),
    }));
    state.register_state_stack.push(state.registers.clone());

    // zero registers and make sure there are enough for dm_function
    state.registers.clear();
    if state.registers.len() < dm_fn.number_used_registers() {
        state
            .registers
            .resize(dm_fn.number_used_registers(), DvmValue::Uninit);
    }

    // push args
    for (i, arg) in args.iter().enumerate() {
        state.registers[i] = arg.clone();
    }

    // run the byte code
    run_byte_code(state, context, &dm_fn.byte_code());

    // restore state
    state.registers = state.register_state_stack.pop().unwrap();
    state.call_stack.pop();

    // get return value
    if let Some(return_register) = dm_fn.return_register() {
        state
            .registers
            .get(return_register as usize)
            .map(|m| m.clone())
            .unwrap()
    } else {
        DvmValue::Void
    }
}

macro_rules! next_8 {
    ( $byte_code: expr, $T: ident ) => {
        $byte_code.next().unwrap() as $T
    };
}

macro_rules! next_32_le {
    ( $byte_code: expr, $T: ident ) => {
        next_8!($byte_code, $T)
            + (next_8!($byte_code, $T) << 8)
            + (next_8!($byte_code, $T) << 16)
            + (next_8!($byte_code, $T) << 24)
    };
}

macro_rules! next_64_le {
    ( $byte_code: expr, $T: ident ) => {
        next_8!($byte_code, $T)
            + (next_8!($byte_code, $T) << 8)
            + (next_8!($byte_code, $T) << 16)
            + (next_8!($byte_code, $T) << 24)
            + (next_8!($byte_code, $T) << 32)
            + (next_8!($byte_code, $T) << 40)
            + (next_8!($byte_code, $T) << 48)
            + (next_8!($byte_code, $T) << 56)
    };
}

macro_rules! next_usize_le {
    ( $byte_code: expr ) => {{
        let mut u: usize = 0;
        for i in 0..std::mem::size_of::<usize>() {
            u += next_8!($byte_code, usize) << (8 * i);
        }
        u
    }};
}

macro_rules! read_array_u8 {
    ( $iter: expr ) => {{
        let arr_length = next_8!($iter, u8);
        let mut arr = Vec::with_capacity(arr_length as usize);
        let mut i = 0;
        while i < arr_length {
            arr.push(next_8!($iter, u8));
            i += 1;
        }
        arr
    }};
}

macro_rules! read_string {
    ( $iter: expr ) => {{
        let str_length = next_32_le!($iter, usize);
        let mut str_raw: Vec<u8> = Vec::new();
        let mut i = 0;
        while i < str_length {
            str_raw.push($iter.next().unwrap());
            i += 1;
        }
        String::from_utf8(str_raw).unwrap()
    }};
}

macro_rules! write_string {
    ( $raw_ptr: expr, $s: expr ) => {
        for (i, b) in $s.bytes().enumerate() {
            unsafe {
                $raw_ptr.add(i).write(b);
            }
        }
    };
}

macro_rules! write_bytes {
    ( $ptr: expr, $offset: expr, $src: expr ) => {
        for (i, &b) in $src.iter().enumerate() {
            unsafe {
                $ptr.add($offset + i).write(b);
            }
        }
    };
}

pub fn run_byte_code(state: &mut DvmState, context: &DvmContext, byte_code: &[u8]) {
    let mut iter = byte_code.iter().cloned();
    while let Some(op_code) = iter.next() {
        match op_code {
            MOV_INT => {
                let target_register = next_8!(iter, usize);
                let operand = next_32_le!(iter, u32) as i32;
                state.registers[target_register] = DvmValue::Int(operand);
            }
            MOV_LONG => {
                let target_register = next_8!(iter, usize);
                let operand = next_64_le!(iter, u64) as i64;
                state.registers[target_register] = DvmValue::Long(operand);
            }
            MOV_DOUBLE => {
                let target_register = next_8!(iter, usize);
                let operand = next_64_le!(iter, u64) as f64;
                state.registers[target_register] = DvmValue::Double(operand);
            }
            MOV_REGISTER => {
                let target_register = next_8!(iter, usize);
                let source_register = next_8!(iter, usize);
                let source_value = state.registers.get(source_register).unwrap();
                state.registers[target_register] = source_value.clone();
            }
            LOAD_OBJECT => {
                let target_register = next_8!(iter, usize);
                let source_register = next_8!(iter, usize);
                let offset = next_usize_le!(iter);
                
                let source_object = state.registers
                    .get(source_register)
                    .unwrap()
                    .expect_object();
                
                let object = unsafe {
                    Rc::from_raw(source_object.data.add(offset).cast::<DmAllocObject>())
                };
                
                state.registers[target_register] = DvmValue::Object(object);
            }
            STORE => {
                let target_register = next_8!(iter, usize);
                let offset = next_usize_le!(iter);
                let source_register = next_8!(iter, usize);

                let target_alloc_object = state
                    .registers
                    .get(target_register)
                    .unwrap()
                    .expect_object();
                
                let source_value = std::mem::replace(&mut state.registers[source_register], DvmValue::Uninit);
                
                match source_value {
                    DvmValue::Int(i) => {
                        write_bytes!(target_alloc_object.data, offset, i.to_ne_bytes());
                    }
                    DvmValue::Byte(b) => unsafe {
                        target_alloc_object.data.add(offset).write(b);
                    },
                    DvmValue::Long(l) => {
                        write_bytes!(target_alloc_object.data, offset, l.to_ne_bytes());
                    }
                    DvmValue::Double(d) => {
                        write_bytes!(target_alloc_object.data, offset, d.to_ne_bytes());
                    }
                    DvmValue::Boolean(b) => unsafe {
                        target_alloc_object.data.add(offset).write(b as u8);
                    },
                    DvmValue::Object(source_object) => {
                        let source_object_ptr = Rc::into_raw(source_object);
                        write_bytes!(
                            target_alloc_object.data,
                            offset,
                            (source_object_ptr as usize).to_ne_bytes()
                        );
                    }
                    DvmValue::USize(u) => {
                        write_bytes!(target_alloc_object.data, offset, u.to_ne_bytes());
                    }
                    DvmValue::Uninit => {
                        panic!("Cannot store DvmValue::Uninit to object.")
                    }
                    DvmValue::Void => {
                        panic!("Cannot store DvmValue::Void to object.")
                    }
                }
            }
            ALLOC => {
                let target_register = next_8!(iter, usize);
                let alloc_size = next_32_le!(iter, usize);
                let impl_name = read_string!(iter);

                let implementation = context
                    .libs
                    .iter()
                    .find_map(|lib| {
                        lib.implementations
                            .iter()
                            .find(|implementation| implementation.fqn == impl_name)
                    })
                    .expect(&format!("Implementation not found: {}", impl_name));

                let dm_alloc_object = DmAllocObject::new(alloc_size, implementation.clone());
                state.registers[target_register] = DvmValue::Object(Rc::new(dm_alloc_object));
            }
            DEALLOC => {
                let target_register = next_8!(iter, usize);
                let dm_alloc_object = state
                    .registers
                    .get(target_register)
                    .unwrap()
                    .expect_object();
                drop(dm_alloc_object); // explicit
                state.registers[target_register] = DvmValue::Uninit;
            }
            MOV_INT_TO => {
                let target_register = next_8!(iter, usize);
                let offset = next_32_le!(iter, usize);
                let operand = next_32_le!(iter, u32) as i32;

                let dm_alloc_object = state
                    .registers
                    .get(target_register)
                    .unwrap()
                    .expect_object();
                write_bytes!(dm_alloc_object.data, offset, operand.to_ne_bytes());
            }
            MOV_LONG_TO => {
                todo!()
            }
            MOV_DOUBLE_TO => {
                todo!()
            }
            MOV_REGISTER_TO => {
                // Moves the value from the source register to the object pointed to by the target
                // register, at the given offset.
                let target_register = next_8!(iter, usize);
                let offset = next_32_le!(iter, usize);
                let source_register = next_8!(iter, usize);

                let target_alloc_object = state
                    .registers
                    .get(target_register)
                    .unwrap()
                    .expect_object();

                let source_value = state.registers.get(source_register).unwrap();
                match source_value {
                    DvmValue::Int(i) => {
                        write_bytes!(target_alloc_object.data, offset, i.to_ne_bytes());
                    }
                    DvmValue::Byte(b) => unsafe {
                        target_alloc_object.data.add(offset).write(*b);
                    },
                    DvmValue::Long(l) => {
                        write_bytes!(target_alloc_object.data, offset, l.to_ne_bytes());
                    }
                    DvmValue::Double(d) => {
                        write_bytes!(target_alloc_object.data, offset, d.to_ne_bytes());
                    }
                    DvmValue::Boolean(b) => unsafe {
                        target_alloc_object.data.add(offset).write(*b as u8);
                    },
                    DvmValue::Object(source_object) => {
                        let source_object_ptr = Rc::into_raw(source_object.clone());
                        write_bytes!(
                            target_alloc_object.data,
                            offset,
                            (source_object_ptr as usize).to_ne_bytes()
                        );
                    }
                    DvmValue::USize(us) => {
                        write_bytes!(target_alloc_object.data, offset, us.to_ne_bytes());
                    }
                    DvmValue::Uninit => {
                        panic!("Cannot move DvmValue::Uninit to object.")
                    }
                    DvmValue::Void => {
                        panic!("Cannot move DvmValue::Void to object.")
                    }
                }
            }
            MOV_CONST => {
                // Decode
                let address_register = next_8!(iter, usize);
                let size_register = next_8!(iter, usize);
                let lib_name = read_string!(iter);
                let const_id = next_32_le!(iter, usize);

                // Get constant
                let Some(lib) = context.libs.iter().find(|lib| lib.name == lib_name) else {
                    panic!("Could not find lib with name: {}", lib_name);
                };
                let constant = lib.constants.get(const_id).unwrap();
                let DmConstant::String(s) = constant;

                // Alloc memory for constant
                let size = s.len();
                let layout = Layout::from_size_align(size, 1).unwrap();
                let raw_pointer = unsafe { alloc(layout) };

                // Move constant to memory
                let bytes = s.as_bytes();
                for i in 0..size {
                    unsafe {
                        raw_pointer.add(i).write(bytes[i]);
                    }
                }

                // Store the pointer metadata in target registers.
                state.registers[address_register] = DvmValue::USize(raw_pointer as usize);
                state.registers[size_register] = DvmValue::USize(size);
            }
            ALLOC_RAW => {
                // Allocates a raw number of bytes, with the number of bytes determined by the value
                // of the source register.
                let address_register = next_8!(iter, usize);
                let source_register = next_8!(iter, usize);
                let size = state.registers.get(source_register).unwrap().expect_long() as usize;
                let layout = Layout::from_size_align(size, 1).unwrap();
                let raw_ptr = unsafe { alloc(layout) };
                state.registers[address_register] = DvmValue::USize(raw_ptr as usize);
            }
            DEALLOC_RAW => {
                // Deallocates a DvmValue::Pointer(DvmPointer) in the target register.
                let address_register = next_8!(iter, usize);
                let size_register = next_8!(iter, usize);
                let address = state
                    .registers
                    .get(address_register)
                    .unwrap()
                    .expect_usize();
                let size = state.registers.get(size_register).unwrap().expect_usize();
                let layout = Layout::from_size_align(size, 1).unwrap();
                unsafe {
                    dealloc(address as *mut u8, layout);
                }
                state.registers[address_register] = DvmValue::Uninit;
                state.registers[size_register] = DvmValue::Uninit;
            }
            PLATFORM_CALL => {
                // Calls a platform function. The result of the platform call is stored in the
                // specified return register.
                let symbol_name = read_string!(iter);
                let return_register = next_8!(iter, usize);
                let arg_registers = read_array_u8!(iter);

                let mut args = Vec::with_capacity(arg_registers.len());
                for arg_register in arg_registers {
                    let value = state.registers.get(arg_register as usize).unwrap();
                    args.push(value.clone());
                }

                let platform_function = context
                    .platform_functions
                    .get(&symbol_name)
                    .unwrap()
                    .clone();
                state
                    .call_stack
                    .push(CallFrame::PlatformCall(CallFrameInfo { fqn: symbol_name }));
                let call_result = platform_function(args, state, context);
                state.registers[return_register] = call_result;
                state.call_stack.pop();
            }
            INVOKE_FN => {
                let symbol_name = read_string!(iter);
                let return_register = next_8!(iter, usize);
                let arg_registers = read_array_u8!(iter);

                let mut args = Vec::with_capacity(arg_registers.len());
                for arg_register in arg_registers {
                    let value = state.registers.get(arg_register as usize).unwrap();
                    args.push(value.clone());
                }

                let dm_fn = context.functions.get(&symbol_name).unwrap();
                let call_result = call_fn(state, context, dm_fn, args);
                state.registers[return_register] = call_result;
            }
            INVOKE_VIRTUAL => {
                let symbol_name = read_string!(iter);
                let return_register = next_8!(iter, usize);
                let arg_registers = read_array_u8!(iter);

                let mut args = Vec::with_capacity(arg_registers.len());
                for arg_register in arg_registers {
                    let value = state.registers.get(arg_register as usize).unwrap();
                    args.push(value.clone());
                }

                let self_obj = args.get(0).unwrap().expect_object();
                let method = self_obj
                    .implementation
                    .get_method(&symbol_name, &self_obj)
                    .expect(&format!("Could not find method: {}", symbol_name));
                let dm_fn = method.dm_fn();

                let call_result = call_fn(state, context, &dm_fn, args);
                state.registers[return_register] = call_result;
            }
            INVOKE_DYNAMIC => {
                unimplemented!("INVOKE_DYNAMIC is not yet supported and may never be.")
            }
            MOV_SIZE_OF => {
                todo!()
            }
            MULTIPLY => {
                todo!()
            }
            op_code => {
                panic!("Invalid op_code: {}", op_code);
            }
        }
    }
}

#[cfg(test)]
mod run_code_tests {
    use crate::vm::dm_type::DmType;
    use crate::vm::object_type::{DmField, DmImplementation};
    use super::*;

    macro_rules! assert_register {
        ( $expected: expr, $state: expr, $register_number: expr ) => {
            assert_eq!(
                $expected,
                $state.registers.get($register_number).unwrap().clone()
            );
        };
    }

    fn setup() -> (DvmState, DvmContext) {
        (DvmState::new(), DvmContext::new())
    }
    
    fn impl_with_object_field() -> DmImplementation {
        let mut dm_impl = DmImplementation::new("ImplWithObjectField", "ImplWithObjectField", None);
        let object_field = DmField::new("object_field", DmType::Object, 0);
        dm_impl.fields.push(object_field);
        dm_impl
    }

    #[test]
    fn mov_1_as_int() {
        let mut code = Vec::new();
        add_mov_int(&mut code, 0, 1);
        let (mut state, context) = setup();
        run_byte_code(&mut state, &context, &code);
        assert_register!(DvmValue::Int(1), state, 0);
    }

    #[test]
    fn move_65535_as_int() {
        let mut code = Vec::new();
        add_mov_int(&mut code, 0, 0xffff);
        let (mut state, context) = setup();
        run_byte_code(&mut state, &context, &code);
        assert_register!(DvmValue::Int(0xffff), state, 0);
    }

    #[test]
    fn move_int_max_as_int() {
        let mut code = Vec::new();
        add_mov_int(&mut code, 0, 0x0fff_ffff);
        let (mut state, context) = setup();
        run_byte_code(&mut state, &context, &code);
        assert_register!(DvmValue::Int(0x0fff_ffff), state, 0);
    }

    #[test]
    fn move_register() {
        let mut code = Vec::new();
        add_mov_int(&mut code, 1, 1);
        add_mov_register(&mut code, 0, 1);
        let (mut state, context) = setup();
        state.registers.resize(2, DvmValue::Uninit);
        run_byte_code(&mut state, &context, &code);
        assert_register!(DvmValue::Int(1), state, 0);
    }
    
    #[test]
    fn load_object() {
        let dummy_impl = impl_with_object_field();
        let dummy_impl_rc = Rc::new(dummy_impl);
        let dummy_impl_object_field = dummy_impl_rc.fields.iter().find(|field| {
            field.name() == "object_field"
        }).unwrap();
        let mut dummy_lib = DmLib::new("dummy");
        dummy_lib.implementations.push(dummy_impl_rc.clone());
        
        let mut code = Vec::new();
        add_alloc(&mut code, 0, dummy_impl_rc.size_in_bytes() as u32, &dummy_impl_rc.fqn);
        add_alloc(&mut code, 1, dummy_impl_rc.size_in_bytes() as u32, &dummy_impl_rc.fqn);
        add_store(&mut code, 0, dummy_impl_object_field.data_offset(), 1);
        add_load_object(&mut code, 2, 0, dummy_impl_object_field.data_offset());
 
        let (mut state, mut context) = setup();
        state.registers.resize(3, DvmValue::Uninit);
        context.load_libs(vec![Rc::new(dummy_lib)]);
        run_byte_code(&mut state, &context, &code);
        
        let referenced_object = state.registers.get(1).unwrap().expect_object();
        assert_register!(DvmValue::Object(referenced_object), state, 2);
    }
}
